create or replace procedure create_obj(obj_type character varying,
	obj_id uuid,
	name_obj character varying,
	comp_id uuid,
	parent_type character varying,
	parent_uuid uuid,
	root_id uuid,
	ins_type_table boolean default true,
	link_type int default 1,
	add_state boolean default false,
	tm timestamp without time zone default (now() at time zone 'utc'))
	LANGUAGE 'plpgsql' as
$$
declare
	maxcount int;
	rowcount int;
	maxchildcount int;
	
begin
	select  "max_children_count"  as result into maxchildcount  from "types"  where lower("type") = lower("parent_type");
	select count("parent") as result into rowcount from "links" 
		where  lower("linking_obj_type") = lower(parent_type) and "linking_obj_id" = parent_uuid and "parent" = link_type;
	if (maxchildcount <> -1 and rowcount >= maxchildcount) then
		raise exception  using errcode = '70014';		
	return;
	end if;
																						 
	select "max_count" as result into maxcount  from "types" where lower("type") = lower("obj_type");
	select count("parent") as result into rowcount from "links" 
		where  lower("linking_obj_type") = lower(parent_type) and "linking_obj_id" = parent_uuid 
		and lower("linked_obj_type") = lower(obj_type) and "parent" = link_type;
	if (maxcount <> -1 and rowcount >= maxcount) then
		raise exception  using errcode = '70015';	
		return;
	end	if;				
	
	if (ins_type_table = true) then 
		execute 'insert into "t_' || obj_type || '" ("id") values ('||quote_literal(obj_id)||') ;';
	end if;
	insert into "objects" (id, "type", "name", "time") values (obj_id, obj_type, name_obj, tm) ;
	if (add_state = true) then 
		insert into "object_states" ("id", "state","data", "time") values (obj_id, decode('ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'hex'),'',tm);
	end if;
	insert into "root_links" ("id", "comp_id", "root_id") values(obj_id,comp_id, root_id);
	insert into links (linked_obj_id, linked_obj_type, linking_obj_id, linking_obj_type, parent, data, time) values
		(obj_id, obj_type, parent_uuid,parent_type, link_type, 0, tm);
	return;
end
$$;


create or replace procedure create_root_obj(obj_type character varying, 
										   name_obj character varying,
										  comp_id_in uuid, uuid_object uuid,
										add_state boolean default false,
										tm timestamp without time zone default (now() at time zone 'utc'))
										language 'plpgsql'
										as
$$
declare
	maxcount int;
	rowcount int;
begin
	if exists (select "linked_obj_type" from "possible_links" where 
			   (lower("linked_obj_type") = lower(obj_type) and "parent" = 1)) then 
		raise exception  using errcode = '70010';		
	end if;		
	select "max_count" as result into maxcount  from "types" where (lower("types"."type") = lower(obj_type));		
	if (maxcount is null)then 
		raise exception  using errcode = '70010';												
	end if;	
	select count("type") as result into rowcount from "root_links"
	inner join  "objects" on "objects"."id" = "root_links"."id" 	
		where "root_links"."comp_id" = "comp_id_in" and lower("objects"."type") = lower(obj_type);
	if ( maxcount <> -1 and  rowcount >= maxcount )then
		raise exception  using errcode = '70015';
	end if;
	if name_obj = '' then	
		select types.name as result into name_obj 
			from "types" where types.type=obj_type;	
	end if;	
	execute 'insert into "t_' || obj_type || '" ("id") values ('||quote_literal(uuid_object)||') ;';	
	insert into "objects" ("id", "type", "name", "time") values (uuid_object, obj_type, name_obj, tm) ;
	if (add_state = true) then 
		insert into "object_states" ("id", "state","data", "time") values (uuid_object, decode('ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff', 'hex'),'', tm);
	end if;	
	insert into "root_links" ("id", "comp_id", "root_id") values(uuid_object,comp_id_in, uuid_object);
  	return;
end
$$;


create or replace function check_change_parent(
	parent_type character varying,
	parent_id uuid,
	child_type character varying,
	child_id uuid)
    returns bool
    language 'plpgsql'
as $body$
declare
	temp_type_parent character varying;
	temp_type_child character varying;
	temp_id_parent uuid;
	temp_id_child uuid;
	parent_comp_id uuid;
	child_comp_id uuid;
	maxcount int;
	rowcount int;
	maxchildcount int;
begin
	select "comp_id" as result into "parent_comp_id"  from "root_links" where "id" = parent_id;
	select "comp_id" as result into child_comp_id  from "root_links" where "id" = child_id;
	if parent_comp_id <> child_comp_id then
		raise exception using errcode = '70005';
	end if;
	if exists (select "parent" from "possible_links" where (lower("linking_obj_type") = lower("parent_type")) 
			and (lower("linked_obj_type")= lower(child_type)) and ("parent" = 1)) then
		if not exists(select * from "objects" where (lower("type") = lower("parent_type")) and ("id" = parent_id)) then
			 raise exception  using errcode = '70005';												   
	   	end if;
	else															  
		raise exception using errcode = '70004';																   
	end if;
																								
	select  "max_children_count"  as result into maxchildcount  from "types"  where lower("type") = lower("parent_type");
	select count("parent") as result into rowcount from "links" 
		where  lower("linking_obj_type") = lower(parent_type) and "linking_obj_id" = parent_id and "parent" = 1;
	if (maxchildcount <> -1 and rowcount >= maxchildcount) then
		raise exception  using errcode = '70014';	
	
	end if;

	select "max_count" as result into maxcount  from "types" where lower("type") = lower("child_type");
	select count("parent") as result into rowcount from "links" 
		where  lower("linking_obj_type") = lower(parent_type) and "linking_obj_id" = parent_id 
		and lower("linked_obj_type") = lower(child_type) and "parent" = 1;
	if (maxcount <> -1 and rowcount >= maxcount) then
		raise exception  using errcode = '70015';
	end	if;																								
	temp_type_child  := parent_type;	
	temp_id_child :=parent_id;	
	temp_type_parent := temp_type_child;
	temp_id_parent := temp_id_child;	
	while exists (select "id" from "root_links" where "id" = temp_id_parent) loop																								
		select  "linking_obj_id","linking_obj_type" into  temp_id_parent, temp_type_parent
			from "links" where "linked_obj_id" =temp_id_child and "linked_obj_type" = temp_type_child  and "parent" = 1 limit 1;
		if 	temp_id_parent = child_id then
			raise exception  using errcode = '70011';
		end if;																						
	    if	temp_id_parent is not null then			
			temp_id_child := temp_id_parent;
			temp_type_child := 	temp_type_parent;																				
		end if;																								
	end loop; 	
	return true;
end;
											 
$body$;

create or replace procedure public.delete_obj( _type character varying,
	_id uuid, _name character varying, _comp_id uuid, _root_id uuid, 
	 _params character varying, _link_type integer,
	 tm timestamp without time zone)
    language 'plpgsql' 
as $body$
declare
	_script varchar= '';
	_count integer=0;
begin
	if exists (select "type" from "objects" where "id"= _id and "type" =_type ) then
		select into _count count(*) from "links" where "linked_obj_id" = _id and
			"linked_obj_type" = lower(_type) and "parent" = 1;
		if _count > 1 then
			raise exception  using errcode = '70015';
		end if;
		if not exists ( select "parent"  from "links" where "linking_obj_id" = _id and
		"linking_obj_type" = lower(_type) and "parent" = _link_type ) then
			delete from "object_states" where "id" = _id;
			with 
			chain_ids AS (select "id" from "link_chains" where "obj_id" = _id),
			delete_data AS (delete from "link_chain_data" where "id" in (select "id" from chain_ids))
			delete from "link_chains" where "id" in (select "id" from chain_ids);			
			delete from "links" where "linking_obj_id" = _id or "linked_obj_id" = _id;
			delete from "root_links" where "id" = _id;
			delete from "objects" where "id" = _id;
			execute 'delete from "t_' || _type || '"where  "id" = '||quote_literal(_id)||';';
			select string_agg(('delete from '|| '"t_' || _type || '_' || "dependent_tables"."add_params_name"
				||'" where "id" = ' || quote_literal(_id) || ' ; '|| e'\n'),'') as result into _script
				from "dependent_tables" where "type" = lower(_type) ;
			if _script is not null then 
				execute _script;
			end if;
			insert into "delete_objects" ("type", "id","params","name","time","comp_id","root_id") values
			(_type, _id,_params,_name, tm,_comp_id,_root_id);
		else
			raise exception  using errcode = '70015';
		end if; 		
	else
		raise exception  using errcode = '70015';
	end if;																  
return;
end
$body$;

create or replace function public.set_link( _linking_type character varying,
	_linking_id uuid, _linked_type character varying,
	_linked_id uuid,_data character varying, 
	_parent integer,
	tm timestamp without time zone)
	returns record 
    language 'plpgsql' 
    volatile 
as $body$
declare
	res record;
begin
	if exists( select "linking_obj_type" from "possible_links" where "linking_obj_type" = _linking_type
				and "linked_obj_type" = _linked_type and "parent" = _parent ) then
		if not exists( select "type" from "objects" where "type" = lower(_linking_type)	and "id" = _linking_id ) then
			raise exception  using errcode = '70005';
		end if;	
		if not exists( select "type" from "objects" where "type" =  lower(_linked_type)	and "id" = _linked_id ) then
			raise exception  using errcode = '70005';
		end if;	
		if not exists( select "parent" from "links" where "linking_obj_type" = lower(_linking_type)
					and "linked_obj_type" = lower(_linked_type)  and "linking_obj_id" = _linking_id 
			  		and "linked_obj_id" = _linked_id and "parent" = _parent ) then
			insert into "links"("linking_obj_type", "linking_obj_id", "linked_obj_type", 
				"linked_obj_id", "parent", "time","data")
				values (_linking_type, _linking_id, _linked_type,_linked_id , _parent, tm,_data);
			select "root_links"."comp_id" as linking_comp_id, rl."comp_id" as linked_comp_id into res from "root_links"
			inner join "root_links" as rl on rl."id" = _linked_id where "root_links"."id" = _linking_id;
		else
			raise exception  using errcode = '70006';
		end if;
	else
		raise exception  using errcode = '70007';
	end if;
	return res;
end
$body$;

drop function if exists public.delete_rights_type() cascade;
drop function if exists public.create_rights_type() cascade;
drop function if exists public.delete_user_type() cascade;

create function public.delete_rights_type()
    returns trigger
    language 'plpgsql'
as $body$
begin
	delete from "user_rights" where "user_rights"."rights_id" =  old."rights_id";
	delete from "rights" where "rights_id"  = old."rights_id";
	return old;
end;
$body$;

create function public.create_rights_type()
    returns trigger
    language 'plpgsql'
as $body$
declare 
	_new_id uuid;
begin
	_new_id = uuid_generate_v4();
	insert into "rights"( "rights_id", "name", "permissions")
	values (_new_id,'rights', '');
	new."rights_id" = _new_id;
	return new;
end;
$body$;

create function public.delete_user_type()
    returns trigger
    language 'plpgsql'
as $body$
begin 
	delete from "user_rights" where "user_rights"."user_id"  in	(select old."id" );
	return old;
end;
$body$;

create or replace procedure public.change_user_rights(_rights_id uuid,_user_type character varying,_user_id uuid)
    language 'plpgsql'
as $body$
begin
	if exists (select  "user_id" from "user_rights" where "user_id" = _user_id and "user_type" = _user_type) then	
		update "user_rights" set "rights_id"= _rights_id where "user_id" = _user_id and "user_type" = _user_type;
	else
		raise exception  using errcode = '70013';	
	end if;
	return;
end
$body$;

CREATE OR REPLACE PROCEDURE public.delete_incident(
	_id uuid)    
    LANGUAGE 'plpgsql'
AS $BODY$
begin
	delete from "incidents" where "id" = _id;
	delete from "incident_message_links" where "id" = _id;
	delete from "incident_object_links" where "id" = _id;
return;
end
$BODY$;

CREATE OR REPLACE FUNCTION public.get_number_object(
	id uuid,
	type character varying)
    RETURNS character varying
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE 
AS $BODY$
DECLARE 
	num  character varying;
begin
 	if not exists (SELECT  col.column_name FROM INFORMATION_SCHEMA."columns" as col 
		where col.column_name = 'number' and col.table_name = 't_'||type ) then	 
	return '';
	end if;
	execute 'select "number" from "t_' || type || '" where "id" = '||quote_literal(id)||'' into num;	
	return num;
end;
$BODY$;

create or replace function public.get_int_value_object(
	id uuid,
	type character varying,
	field character varying)
    returns character varying
    language 'plpgsql'

    cost 100
    volatile 
as $$
declare 
	val  integer;
begin
 	if not exists (select  col.column_name from information_schema."columns" as col 
		where col.column_name = field and col.table_name = 't_'||type ) then	 
	return '';
	end if;
	execute 'select "'||field||'" from "t_' || type || '" where "id" = '||quote_literal(id)||'' into val;	
	return val;
end;
$$;

create or replace function public.get_history_by_time(
	time_begin timestamp without time zone,
	time_end timestamp without time zone)
    returns table(event_id uuid, status integer, type character varying, id uuid, action integer, dtime timestamp without time zone, params character varying, user_type character varying, user_id uuid, comp_id uuid, msg_flags integer, comment character varying) 
    language 'plpgsql'

    cost 100
    volatile 
    rows 1000
as $$
declare 
	_row record;
	_sql varchar;
begin
	for _row in select date_part('month', gs) ||'_'|| date_part('year', gs) as name_table from generate_series(time_begin, time_end + interval '1 month', interval '1 month') as gs 
	loop
		if exists (select table_name from information_schema.tables  where  table_schema = 'history' and table_name = _row.name_table) then 
			if  _sql is null then
				_sql := 'select * from  history."' ||_row.name_table || '"' ;
			else
				_sql := _sql || ' union ' || 'select * from  history."' ||_row.name_table|| '"';
			end if;
		end if;			
	 
   	end loop;
	if  _sql is not null then
		raise notice '%', quote_ident(_sql);
		return query execute _sql;
	end if;
		
end;
$$;